class Status:
    SKIPPED = "SKIPPED"
    FAILED = "FAILED"
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    COMPLETED = "COMPLETED"
    RETRYING = "RETRYING"


class Queue:
    CORE = "core"


class Task:
    SEND_VERSION = "send_version_to_core"



from iso639 import Lang


def lang_to_iso639_1(lang: str, unk_code: str = "unk") -> str:
    """ Converts language into ISO-639-1 standard. 
    Input can be any language code in a valid ISO-639 
    standard or even a full name of the language, 
    e.g. "Estonian".
    
    Parameters
    -----------
    lang: str
        Language code in any valid ISO-639 standard.
        
    unk_code: str
        Code to return incase of invalid/unsupported
        input language.
        
    Returns
    -------
    Language code in ISO-639-1 standard.
    """
    try:
        lg = Lang(lang)
        iso_639_1_lang = lg.pt1
    except:
        iso_639_1_lang = unk_code
    return iso_639_1_lang


def lang_to_iso639_2(lang: str, unk_code: str = "unk") -> str:
    """ Converts language into ISO-639-2 standard. 
    Input can be any language code in a valid ISO-639 
    standard or even a full name of the language, 
    e.g. "Estonian".
    
    Parameters
    -----------
    lang: str
        Language code in any valid ISO-639 standard.
        
    unk_code: str
        Code to return incase of invalid/unsupported
        input language.
        
    Returns
    -------
    Language code in ISO-639-2 standard.
    """
    try:
        lg = Lang(lang)
        # NB! uses bibliographic identifier (e.g. "de" -> "ger")
        # opposed to terminological identifier ("de" -> "deu").
        # This can be changed by replaving lg.pt2b -> lg.pt2t
        iso_639_2_lang = lg.pt2b
    except:
        iso_639_2_lang = unk_code
    return iso_639_2_lang


def lang_to_iso639_3(lang: str, unk_code: str = "unk") -> str:
    """ Converts language into ISO-639-3 standard. 
    Input can be any language code in a valid ISO-639 
    standard or even a full name of the language, 
    e.g. "Estonian".
    
    Parameters
    -----------
    lang: str
        Language code in any valid ISO-639 standard.
        unk_code: str
        
    Code to return incase of invalid/unsupported
        input language.
        
    Returns
    -------
    str
        Language code in ISO-639-3 standard.
    """
    try:
        lg = Lang(lang)
        iso_639_3_lang = lg.pt3
    except:
        iso_639_3_lang = unk_code
    return iso_639_3_lang


def ratio_to_percentage(ratio: float) -> str:
    """ Converts ratio to corresponding percentage.
    
    Parameters
    -----------
    ratio: float
        Float in range [0,1]
    
    Returns
    --------
    str
        Percentage corresponding to the float.
    
    """
    percentage = f"{int(ratio*100)}%"
    return percentage



from rara_tools.normalizers.viaf import VIAFRecord, VIAFClient


def test_fetch_clusters_by_id_list():
    viaf_ids = ["7432247", "456"]
    client = VIAFClient()
    
    results = client.fetch_viaf_clusters(viaf_ids)
    assert len(results) == 2
    assert results["456"] == {}
    assert len(results["7432247"]) > 0
    
    
def test_fetch_viaf_results_for_normalizer():
    viaf_ids = ["7432247", "456"]
    client = VIAFClient()
    
    results = client.get_normalized_data(viaf_ids)
    assert len(results) == 2


from typing import Optional

import requests

from .exceptions import TaskReporterException


class TaskReporter:
    """A class to manage task reporting to Core API.
    Requires:
    * running instance of Core API,
    * Core API access token with sufficient privileges.
    """

    def __init__(
            self,
            api_url: str,
            api_token: str,
            api_timeout: Optional[int] = 10,
    ):
        # Remove possible trailing / from API url.
        self.api_url = api_url.rstrip("/")
        self.api_headers = {
            "ContentType": "application/json",
            "Authorization": f"Token {api_token}"
        }
        self.api_timeout = api_timeout

    def check(self):
        """Checks the connection to Core API.
        This includes both connection & auth errors.
        :return: Boolean - true for alive, false for dead.
        """
        health_url = f"{self.api_url}/health"
        response = requests.get(health_url, headers=self.api_headers, timeout=self.api_timeout)
        if response.status_code == 200:
            return True
        return False

    def _perform_patch(self, content: dict, task_id: int, tasks_endpoint: str = "tasks"):
        url = f"{self.api_url}/{tasks_endpoint}/{task_id}/"
        try:
            return requests.patch(
                url,
                headers=self.api_headers,
                json=content,
                timeout=self.api_timeout
            )
        except Exception as e:
            raise TaskReporterException(f"Error patching document: {e}")

    def _perform_status_update(self, content: dict, task_id: int, tasks_endpoint: str = "tasks", action_endpoint: str = "update_status"):
        url = f"{self.api_url}/{tasks_endpoint}/{task_id}/{action_endpoint}/"
        try:
            return requests.post(
                url,
                headers=self.api_headers,
                json=content,
                timeout=self.api_timeout
            )
        except Exception as e:
            raise TaskReporterException(f"Error patching document: {e}")

    def report_results(self, content: dict, task_id: int):
        """Reports task results to Core API.
        :param: content dict: Keys & values to update in Task model.
        :param: task_id int: Task ID of the object to update.
        :return: Response object.
        """
        response = self._perform_patch(content, task_id)
        if response.status_code not in (200, 204):
            raise TaskReporterException(f"Error code from Core API: {response.status_code}")
        return response

    def update_status(self, key: str, task_id: int, **status_options):
        """Updates the status of sub-tasks to Core API.
        :param: key: Since task statuses are Many-to-Many a normal patch won't work, hence we filter the status we want through the key.
        :param: status_options: Keys & values to update in the status model.
        :return: Response object.
        """
        content = {"key": key, **status_options}
        response = self._perform_status_update(content, task_id)
        if response.status_code not in (200, 204):
            raise TaskReporterException(f"Error code from Core API: {response.status_code}")
        return response



from rara_tools.normalizers import BibRecordNormalizer, AuthoritiesRecordNormalizer
from tests.test_utils import (get_linker_res_example, get_formatted_sierra_response,
                              check_record_tags_sorted, check_no_dupe_tag_values, check_record_tags_have_values)


from pymarc import Record

import os

TEST_LEVEL = os.getenv("TEST_LEVEL", "unit")

EMPTY_SIERRA_RECORDS = [
    {
        "sierraID": "1",
        "leader": "00000nz  a2200000n  4500",
        "fields": []
    },
]

REQUIRED_FIELDS = ["667", "925"]  # always included after normalization
MOCK_LINKER_ONE_FOUND = get_linker_res_example(
    "oneFound.json")
MOCK_LINKER_MULTIPLE_FOUND = get_linker_res_example(
    "multipleFound.json")
MOCK_LINKER_NOT_FOUND = get_linker_res_example(
    "notFound.json")


def test_normalizers_OK():
    """ Test field editing logic & internals """

    linking_results = [MOCK_LINKER_ONE_FOUND,
                       MOCK_LINKER_MULTIPLE_FOUND]

    test_sierra_data = get_formatted_sierra_response("authorities.json")

    normalizer = AuthoritiesRecordNormalizer(
        linking_results=linking_results,
        sierra_data=test_sierra_data,
    )

    assert len(normalizer.records_extra_data) == len(normalizer.data)

    normalizer = BibRecordNormalizer(
        linking_results=linking_results,
        sierra_data=test_sierra_data,
    )
    assert len(normalizer.records_extra_data) == len(normalizer.data)

    data = [
        {
            "sierraID": "1",
            "leader": "00000nz  a2200000n  4500",
            "fields": [
                {
                    "667": {
                        "ind1": " ",
                        "ind2": " ",
                        "subfields": [
                            {
                                "a": "Val"
                            }
                        ]
                    }
                },
            ]
        },
    ]

    # default behavior - added if not in record &
    normalizer = AuthoritiesRecordNormalizer(
        sierra_data=data,
        ALLOW_EDIT_FIELDS=[],
        REPEATABLE_FIELDS=[],
    )
    for r in normalizer:
        assert r.get_fields("667")[0].get_subfields("a")[0] == "Val"

    # not edited if exists
    normalizer = AuthoritiesRecordNormalizer(
        sierra_data=data,
        ALLOW_EDIT_FIELDS=[],
        REPEATABLE_FIELDS=[]
    )
    for r in normalizer:
        assert r.get_fields("667")[0].get_subfields("a")[0] == "Val"

    # allow repeatable, new field will be added
    normalizer = AuthoritiesRecordNormalizer(
        sierra_data=data,
        ALLOW_EDIT_FIELDS=[],
        REPEATABLE_FIELDS=["667"]
    )
    for r in normalizer:
        fields_667 = r.get_fields("667")
        assert len(fields_667) == 2
        assert fields_667[0].get_subfields("a")[0] == "Val"
        assert fields_667[1].get_subfields("a")[0] == "Muudetud AI poolt"

    # allow editing, field will be edited
    normalizer = AuthoritiesRecordNormalizer(
        sierra_data=data,
        ALLOW_EDIT_FIELDS=["667"],
        REPEATABLE_FIELDS=[]
    )
    for r in normalizer:
        fields_667 = r.get_fields("667")
        assert len(fields_667) == 1
        assert fields_667[0].get_subfields("a")[0] == "Muudetud AI poolt"


def validate_bibrecord_normalized(record: Record, has_viaf_data=False):
    # source notes
    assert record.get_fields("667")[0].get_subfields("a")[
        0] == "Muudetud AI poolt"


def validate_authorities_record_normalized(record: Record, has_viaf_data=False):

    field_667 = record.get_fields("667")[0].get_subfields("a")[0]
    assert field_667 == "Muudetud AI poolt" or field_667 == "Loodud AI poolt"

    field_040_subfields = record.get_fields("040")[0]

    # check that a, b & c subfields have values (can have default or unique)
    assert len(field_040_subfields.get_subfields("a")) > 0
    assert len(field_040_subfields.get_subfields("b")) > 0
    assert len(field_040_subfields.get_subfields("c")) > 0

    # check that 008 field has a value of length 40
    field_008 = record.get_fields("008")[0].data
    assert len(field_008) == 40

    if has_viaf_data:
        field_043 = record.get_fields("043")[0].get_subfields(
            "c")[0]  # check that 043 has subfield c with value "ee"
        assert field_043 == "ee"

        field_024 = record.get_fields("024")
        for f in field_024:
            assert len(f.get_subfields("0")) > 0  # VIAF url

        field_046 = record.get_fields("046")[0]
        assert len(field_046.get_subfields("f")) > 0  # birth date
        assert len(field_046.get_subfields("g")) > 0  # death date
        # assert len(field_046.get_subfields("s")) > 0 # activity start
        # assert len(field_046.get_subfields("t")) > 0 # activity end


def test_missing_fields_created_bibrecord_normalization():
    linking_results = [MOCK_LINKER_ONE_FOUND]

    normalizer_entities_only = BibRecordNormalizer(
        linking_results=linking_results,
    )

    normalizer_sierra_data_only = BibRecordNormalizer(
        sierra_data=EMPTY_SIERRA_RECORDS,
    )

    for record in normalizer_entities_only:
        check_record_tags_have_values(
            record, ["008", "046", "245",  # Sierra related, always with bibs
                     "035",  "100",  # VIAf enriched
                     ] + REQUIRED_FIELDS
        )
        validate_bibrecord_normalized(record, has_viaf_data=True)

    for record in normalizer_sierra_data_only:
        check_record_tags_have_values(
            record, ["008", "046", "245",  # Sierra related, always with bibs
                     ] + REQUIRED_FIELDS)
        validate_bibrecord_normalized(record)


def test_missing_fields_created_authorities_normalization():

    linking_results = [MOCK_LINKER_ONE_FOUND]

    normalizer_entities_only = AuthoritiesRecordNormalizer(
        linking_results=linking_results,  # find one match
    )

    normalizer_sierra_data_only = AuthoritiesRecordNormalizer(
        sierra_data=EMPTY_SIERRA_RECORDS,
    )

    for r in normalizer_entities_only:
        check_record_tags_have_values(r, ["008", "040",  # SIERRA related
                                          "024", "043", "046"  # VIAF enriched
                                          ] + REQUIRED_FIELDS)

        validate_authorities_record_normalized(r, True)

    for r in normalizer_sierra_data_only:
        check_record_tags_have_values(
            r, ["040"] + REQUIRED_FIELDS)
        validate_authorities_record_normalized(r)


def test_normalized_fields_sorted():

    unsorted_bibdata = [
        {
            "sierraID": "1",
            "leader": "00000nz  a2200000n  4500",
            "fields": [
                {
                        "035": {
                            "ind1": " ",
                            "ind2": " ",
                            "subfields": [
                                {
                                    "a": "(ErESTER)<1>"
                                }
                            ]
                        }
                },
                {
                    "008": "220805|||aznnnaabn          || |||      nz n  "
                },
                {
                    "046": {
                        "ind1": " ",
                        "ind2": " ",
                        "subfields": [
                            {
                                "k": "1912"
                            }

                        ]
                    }
                },
            ]
        }
    ]

    normalizers = (BibRecordNormalizer, AuthoritiesRecordNormalizer)

    for normalizer in normalizers:
        normalizer = normalizer(
            linking_results=[],
            sierra_data=unsorted_bibdata
        )

        for r in normalizer:
            check_no_dupe_tag_values(r)
            check_record_tags_sorted(r)


def test_authority_normrecord_found_in_es_and_normalized():
    """ KATA elastic normkirjete seast leitakse 1 vaste & normaliseerija täiendab leitud normkirjet VIAF infoga.
        - valideeri normaliseerimise mapping, mis autori tabelis. Täiendatud väljad ja VIAFist info
        - Valideeri märge lisatud (TODO) """
    # Presume, author name identified and sent to linker
    linker_res = get_linker_res_example(
        "oneFound.json")  # single result
    linking_results = [linker_res]

    # 1 result found
    normalizer = AuthoritiesRecordNormalizer(
        linking_results=linking_results
    )

    data = normalizer.data

    assert len(data) == 1

    for r in normalizer:
        check_record_tags_have_values(r, ["040"] + REQUIRED_FIELDS)
        validate_authorities_record_normalized(r, has_viaf_data=True)


def test_authority_normrecord_not_found_in_es_and_viaf():
    """KATA elastic normkirjete seast vastet ei leitud & linkija sooritab VIAFisse otsingu
        - Üks vaste leiti - luuakse uus normkirje
        - Ei leitud ühtegi vastet, või on leitud vasteid mitu - AI tuvastatud info põhjal uue kirje loomine(TODO)
    """
    linker_res = get_linker_res_example(
        "oneFound.json")
    linking_results = [linker_res]

    normalizer = AuthoritiesRecordNormalizer(
        linking_results=linking_results)

    data = normalizer.data

    assert len(data) == 1  # should create new normalized record

    # Entities not found, es & VIAF
    linking_results = [MOCK_LINKER_NOT_FOUND]
    normalizer = AuthoritiesRecordNormalizer(linking_results=linking_results)
    data = normalizer.data
    # should create new normalized record in the future, none for now
    assert len(data) == 0

    linker_res = get_linker_res_example(
        "multipleFound.json")
    linking_results = [linker_res]
    normalizer = AuthoritiesRecordNormalizer(linking_results=linking_results)
    data = normalizer.data
    # should create new normalized record in the future, none for now
    assert len(data) == 0


def test_matching_sierra_record_viaf_id_found():
    """normkirjelt leitakse VIAF ID, vajadusel normi asukoht, kus see ID sisaldub."""
    pass


def test_matching_sierra_record_viaf_id_not_found():
    """kirjelt VIAF IDd ei leitud, soorita otsing VIAFi pihta, et leida _vastutav isik_?. Loo uus vastavalt otsingu tulemusele."""
    pass


def test_authorities_normalizer_checks():
    """
    - kontrolli kas tuvastatud nimi on SIERRAst leitud vaste 1XX, 4XX väljadel. Kui pole, siis lisa 4XX väljale.
    - kontrolli, kas VIAF andmete nimekujud on normkandes olemas. Kui pole, lisa need 4XX väljale.
    - Kontrolli, kas VIAF kandes on sünni ja surma daatumid ja kas need klapivad normkandes olevaga. Kui pole, siis liiguta normkandest kogu 1XX väli 4XX väljale. Seejärel loo uute daatumitega 1XX väli.
    - Kontrolli, et väljal 046 olevad daatumid klapiksid just 1xx väljale lisatuga. Kui andmeid muudeti, siis märgi, et baasis on normkanne muutunud
    """
    pass



from typing import List, NoReturn, Tuple
from abc import abstractmethod
from pymarc.record import Record
from rara_tools.constants.parsers import GeneralMarcIDs


class BaseRecord:
    """ Implements general logic of parsing MARC files.
    """
    def __init__(self, record: Record, add_variations: bool = False) -> NoReturn:
        """ Initializes BaseRecord object.

        Parameters
        -----------
        record: Record
            pymarc.record.Record objectself.
        add_variations: bool
            If enabled, constructs an additional variations field, which
            combines the content of multiple fields + adds some generated
            variations. If the output is uploaded into Elastic and used
            via rara-norm-linker, it is necessary to enable this.
        """
        self.add_variations: bool = add_variations
        self.__record_mrc: Record = record
        self.__record_dict: dict = record.as_dict()["fields"]

        self.__id_field_id: List[str] = GeneralMarcIDs.ID
        self.__id_source_field_id: List[str] = GeneralMarcIDs.ID_SOURCE

        self.__identifier: str = ""
        self.__identifier_source: str = ""

    def get_values(self,
            marc_ids: List[str],
            subfield_id: str | List[str] = "",
            ind1: str = " ",
            ind2: str = " ",
            subfield_restriction: Tuple[str, str] = (),
            subfield_to_ignore: str | None = None
        ) -> List[str] | List[dict]:
        values = []

        for field in self.dict_record:
            field_id = list(field.keys())[0]
            if field_id in marc_ids:
                # TODO: ind1!
                if not subfield_id:
                    values.append(field[field_id])
                else:
                    if field[field_id]["ind2"] == ind2:
                        subfields = field[field_id]["subfields"]
                        subfield_tuples = [list(subfield.items())[0] for subfield in subfields]
                        subfield_keys = [list(subfield.keys())[0] for subfield in subfields]
                        if subfield_restriction and subfield_restriction not in subfield_tuples:
                            continue
                        if subfield_to_ignore and subfield_to_ignore in subfield_keys:
                            continue
                        _value = {}
                        for subfield in subfields:
                            _subfield_id = list(subfield.keys())[0]
                            if isinstance(subfield_id, str):
                                if _subfield_id == subfield_id:
                                    value = subfield[_subfield_id]
                                    values.append(value.strip())
                            elif isinstance(subfield_id, list):
                                if _subfield_id in subfield_id:
                                    value = subfield[_subfield_id]
                                    _value[_subfield_id] = value
                        if isinstance(subfield_id, list):
                            values.append(_value)

        return values

    def _clean_value(self, value: str) -> str:
        cleaned_value = value.strip("., ")
        return cleaned_value

    def _merge_and_clean(self, value: dict, keys: List[str]) -> str:
        _merged = []
        for key in keys:
            _value = self._clean_value(value.get(key, ""))
            if _value:
                _merged.append(_value)
        merged = " ".join(_merged)
        return merged


    @property
    def identifier(self) -> str:
        if not self.__identifier:
            values = self.get_values(marc_ids=self.__id_field_id)
            self.__identifier = values[0] if values else ""
        return self.__identifier

    @property
    def identifier_source(self) -> str:
        if not self.__identifier_source:
            values = self.get_values(marc_ids=self.__id_source_field_id)
            self.__identifier_source = values[0] if values else ""
        return self.__identifier_source

    @property
    def marc_record(self) -> Record:
        return self.__record_mrc

    @property
    def marc_json_record(self) -> dict:
        return self.marc_record.as_dict()

    @property
    def dict_record(self) -> Record:
        return self.__record_dict



import requests
import json
from typing import List
from collections import defaultdict

import logging
logger = logging.getLogger(__name__)


class VIAFRecord:
    def __init__(self,
                 record: dict,
                 allowed_sources: List[str] = [
                     "LC", "DNB", "LNB", "NLL", "ERRR", "J9U"]
                 ):
        self.__record: dict = record
        self.__record_data: dict = {}
        self.__allowed_sources: List[str] = allowed_sources
        self.__viaf_id: int = None
        self.__name_variations: List[str] = []
        self.__birth_date: str = None
        self.__death_date: str = None
        self.__occupations: List[str] = []
        self.__all_fields: dict = {}
        self.__nationality: str = ""
        self.__has_isni: bool = False
        self.__author: str = ""
        self.__author_type: str = None
        self.__has_isni: str = ""
        self.__activity_start: str = None
        self.__activity_end: str = None

    @property
    def author(self) -> str:
        if not self.__author:
            self.__author = self.record_data.get(
                "mainHeading", {}).get("text", "")

    @property
    def author_type(self) -> str:
        """type of name (personal, corporate, title, etc)"""
        if not self.__author_type:
            self.__author_type = self.record_data.get("nameType")

    @property
    def viaf_id(self) -> int:
        if not self.__viaf_id:
            self.__viaf_id = self.record_data.get("viafID", "")
        return self.__viaf_id

    @property
    def has_isni(self) -> bool:
        return bool(self.record_data.get("isni", ""))

    def __get_data(self, field_name: str) -> List[str]:
        entries = self.record_data.get(field_name, {}).get("data", [])

        data = []
        for entry in entries:
            sources = entry.get("sources", {}).get("s", [])
            if set(self.__allowed_sources).intersection(set(sources)):
                data.append(entry.get("text", ""))
        return data

    @property
    def record_data(self) -> dict:
        if not self.__record_data:
            try:
                self.__record_data = self.__record["queryResult"]
            except:
                self.__record_data = self.__record["recordData"]["VIAFCluster"]

        return self.__record_data

    @property
    def name_variations(self) -> List[str]:
        if not self.__name_variations:
            self.__name_variations = self.__get_data("mainHeadings")
        return self.__name_variations

    @property
    def birth_date(self) -> str:
        if not self.__birth_date:
            self.__birth_date = self.record_data.get("birthDate", None)
        return self.__birth_date

    @property
    def death_date(self) -> str:
        if not self.__death_date:
            self.__death_date = self.record_data.get("deathDate", None)
        return self.__death_date

    @property
    def occupations(self) -> List[str]:
        if not self.__occupations:
            self.__occupations = self.__get_data("occupation")
        return self.__occupations

    @property
    def activity_start(self) -> str:
        if not self.__birth_date:
            self.__birth_date = self.record_data.get("activityStart", None)
        return self.__birth_date

    @property
    def activity_end(self) -> str:
        if not self.__death_date:
            self.__death_date = self.record_data.get("activityEnd", None)
        return self.__death_date

    @property
    def nationality(self) -> str:
        if not self.__nationality:
            nationalities = self.__get_data("nationalityOfEntity")
            nationalities_dict = defaultdict(int)
            for n in nationalities:
                nationalities_dict[n.lower()] += 1
            if nationalities:
                self.__nationality = sorted(
                    nationalities_dict.items(), key=lambda x: x[1], reverse=True)[0][0]
        return self.__nationality

    @property
    def all_fields(self) -> dict:
        if not self.__all_fields:
            self.__all_fields = {
                "viaf_id": self.viaf_id,
                "name_variations": self.name_variations,
                "birth_date": self.birth_date,
                "death_date": self.death_date,
                "occupations": self.occupations,
                "nationality": self.nationality,
                "activity_start": self.activity_start,
                "activity_end": self.activity_end,
                "has_isni": self.has_isni,
                "author": self.author
            }
        return self.__all_fields


class VIAFClient:
    def __init__(self, viaf_api_url: str = "https://viaf.org/api"):
        self.root_url = viaf_api_url.strip("/")
        self.record_url = f"{self.root_url}/cluster-record"
        self.search_url = f"{self.root_url}/search"
        self.headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }

    def _send_request(self, url: str, data: dict) -> dict:
        return requests.post(url, data=json.dumps(data), headers=self.headers)

    def get_records_by_search_term(self,
                                   search_term: str,
                                   index: str = "viaf",
                                   field: str = "local.names",
                                   page_index: int = 0,
                                   page_size: int = 50
                                   ) -> dict:
        data = {
            "reqValues": {
                "field": field,
                "index": index,
                "searchTerms": search_term
            },
            "meta": {
                "env": "prod",
                "pageIndex": page_index,
                "pageSize": page_size
            }
        }
        response = self._send_request(url=self.search_url, data=data)
        return response

    def get_records_by_viaf_id(self, record_id: str) -> dict:
        data = {
            "reqValues": {
                "recordId": str(record_id)
            }
        }
        response = self._send_request(url=self.record_url, data=data)

        return response

    def fetch_viaf_clusters(self, viaf_ids):

        results = {}

        for viaf_id in viaf_ids:
            try:
                response = self.get_records_by_viaf_id(viaf_id)
                response.raise_for_status()
                results[viaf_id] = response.json()
            except requests.exceptions.RequestException as e:
                logger.error(f"Error fetching VIAF record {viaf_id}: {e}")
                results[viaf_id] = {}

        return results

    def get_normalized_data(self, record_ids: List[str]) -> List[VIAFRecord]:
        """ Fetch data required for normalization from VIAF. """
        response = self.fetch_viaf_clusters(record_ids)
        return [VIAFRecord(response[record_id]) for record_id in record_ids]



import json
import os
import time
import uuid
from time import sleep

import pytest
from rara_tools.elastic import KataElastic

with open("./tests/test_data/elastic_docs.json") as fh:
    TEST_DOCUMENTS = json.load(fh)

es_url = os.getenv("ELASTIC_TEST_URL", "http://localhost:9200")
ELASTIC = KataElastic(es_url)
ELASTIC_BAD = KataElastic("http://locallost:9012")
TEST_INDEX_NAME = "tools_testing_index"
TEST_DOCUMENT_ID = None
TEST_DOCUMENT_INDEX = None
PARENT_ID = uuid.uuid4().hex


@pytest.mark.order(1)
def test_index_creation():
    """ Tests if index created and documents indexed.
    """
    # Create test index
    created = ELASTIC.create_index(TEST_INDEX_NAME)
    assert created["acknowledged"] is True
    time.sleep(2)

@pytest.mark.order(2)
def test_check():
    """Tests health check method.
    """
    assert ELASTIC.check() is True
    # test bad connection
    assert ELASTIC_BAD.check() is False


@pytest.mark.order(2)
def test_creating_index_again():
    """
    Test to see that running the function for index generation doesn't trigger errors
    on duplicates.
    """
    # Create test index
    created = ELASTIC.create_index(TEST_INDEX_NAME)
    assert created is None


@pytest.mark.order(3)
def test_adding_mapping_to_index():
    """Test adding mapping to an index"""
    schema = ELASTIC.generate_mapping()
    result = ELASTIC.add_mapping(TEST_INDEX_NAME, schema)
    assert result["acknowledged"] is True
    # Test adding the mapping again doesn't create errors.
    result = ELASTIC.add_mapping(TEST_INDEX_NAME, schema)
    assert result["acknowledged"] is True


@pytest.mark.order(4)
def test_document_addition():
    # Add test documents
    for document in TEST_DOCUMENTS:
        indexed = ELASTIC.index_document(TEST_INDEX_NAME, document)
        assert indexed["result"] == "created"
    # let it index
    sleep(1)


@pytest.mark.order(5)
def test_bulk_indexing_documents_cause_rollover():
    data = [{"start_page": number, "sequence_nr": 1, "end_page": number, "parent_id": PARENT_ID} for number in range(10)]
    chunks = [data[i:i + 3] for i in range(0, len(data), 3)]
    for chunk in chunks:
        success, errors = ELASTIC.bulk_index(chunk, TEST_INDEX_NAME, rollover_limit=3, refresh="wait_for")
        assert success is 3 or success is 1

    created_indices = ELASTIC.elasticsearch.indices.get(index=f"{TEST_INDEX_NAME}-*", expand_wildcards="open").body
    assert len(created_indices) == 4


@pytest.mark.order(6)
def test_bulk_indexing_and_document_fetch():
    """
    Test that the whole process of indexing a bunch of different texts and then the retrieval
    of only the requested documents works as intended.
    """
    success, errors = ELASTIC.bulk_index(TEST_DOCUMENTS, TEST_INDEX_NAME, rollover_limit=3, refresh="wait_for")

    # Test the integrity of the limiting query.
    result = ELASTIC.get_documents_by_key(TEST_INDEX_NAME, "foo")
    assert len(result) == 2
    result = ELASTIC.get_documents_by_key(TEST_INDEX_NAME, "bar")
    global TEST_DOCUMENT_ID
    global TEST_DOCUMENT_INDEX
    TEST_DOCUMENT_ID = result[0].meta.id
    TEST_DOCUMENT_INDEX = result[0].meta.index
    assert len(result) == 1
    result = ELASTIC.get_documents_by_key(TEST_INDEX_NAME, "loll")
    assert len(result) == 0

    # Check that sorting works as expected.
    results = ELASTIC.get_documents_by_key(TEST_INDEX_NAME, PARENT_ID)
    for index, document in enumerate(results):
        assert document.start_page == index


@pytest.mark.order(7)
def test_document_deleting():
    """
    Tests deleting a document from index.
    """
    deleted = ELASTIC.delete_document(TEST_DOCUMENT_INDEX, TEST_DOCUMENT_ID)
    assert deleted["result"] == "deleted"
    sleep(1)
    # check if document was actually deleted
    result = ELASTIC.get_documents_by_key(TEST_INDEX_NAME, "bar")
    assert len(result) == 0


@pytest.mark.order(8)
def test_index_deleting():
    """
    Tests deleting index. We delete the test index now.
    """
    deleted = ELASTIC.delete_index(TEST_INDEX_NAME)
    for i in range(10):
        ELASTIC.delete_index(f"{TEST_INDEX_NAME}-{i}")
    assert deleted["acknowledged"] is True



from typing import Any, Dict, Iterator, Optional, List

import elasticsearch_dsl
from elastic_transport import ObjectApiResponse
from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk
from elasticsearch_dsl import Index
from elasticsearch_dsl.response import Response

from .decorators import _elastic_connection


class KataElastic:
    """A class to manage all required Elasticsearch operations for Kata.
    """

    TYPE_MAPPING = {
        "keyword": elasticsearch_dsl.Keyword,
        "text": elasticsearch_dsl.Text,
        "float": elasticsearch_dsl.Float,
        "integer": elasticsearch_dsl.Integer,
        "date": elasticsearch_dsl.Date,
    }

    DEFAULT_MAPPING = {
        "text": "keyword",
        "parent_id": "keyword",
        "text_quality": "float",
        "n_chars": "integer",
        "n_words": "integer",
        "language": "keyword",
        "end_page": "integer",
        "start_page": "integer",
        "sequence_nr": "integer",
        "section_title": "keyword",
        "section_type": "keyword",
        "section_meta": "keyword",
    }

    def __init__(self, elasticsearch_url: str, timeout: Optional[int] = None):
        self.timeout = timeout
        self.elasticsearch_url = elasticsearch_url
        self.elasticsearch = Elasticsearch(self.elasticsearch_url, request_timeout=self.timeout)

    def _produce_rollover_index(self, index_prefix: str, rollover_limit: int) -> str:
        indices = self.elasticsearch.indices.get(index=f"{index_prefix}-*", expand_wildcards="open")
        sorted_indices = sorted([(k, v["settings"]["index"]["creation_date"]) for k, v in indices.items()], key=lambda x: x[1], reverse=True)
        sorted_indices = [i[0] for i in sorted_indices]

        # new index name if none exist
        if not len(sorted_indices):
            last_index_name = f"{index_prefix}-0"
            last_index_count = 0
        else:
            last_index_name = sorted_indices[0]
            last_index_count = self.elasticsearch.count(index=last_index_name)["count"]
        # check the size of the last index of the pipeline
        if last_index_count >= rollover_limit:
            new_index_number = int(last_index_name[-1]) + 1
            last_index_name = f"{index_prefix}-{new_index_number}"

        return last_index_name

    @_elastic_connection
    def check(self) -> bool:
        """Checks Elasticsearch connection.
        :return: bool: Elasticsearch alive or dead.
        """
        if self.elasticsearch.ping():
            return True
        return False

    def generate_mapping(self, schema: dict | None = None) -> dict:
        mapping_dsl = elasticsearch_dsl.Mapping()
        mapping = schema or self.DEFAULT_MAPPING
        for field_name, field_type in mapping.items():
            if field_type in self.TYPE_MAPPING:
                # We instantiate the class stored in the type mapping.
                mapping_dsl.field(field_name, self.TYPE_MAPPING[field_type]())
        return mapping_dsl.to_dict()

    @_elastic_connection
    def add_mapping(self, index_name: str, schema: dict):
        index = Index(name=index_name)
        return index.put_mapping(body=schema, using=self.elasticsearch)
    
    
    @_elastic_connection
    def add_vector_mapping(
            self, 
            index_name: str, 
            field: str, 
            schema: Optional[dict] = None, 
            dims: int = 1024
    ) -> dict:
        vector_mapping = {
            "properties": {
                field: {
                    "type": "dense_vector", 
                    "dims": dims
                }
            }
        }
        mapping = schema or vector_mapping
        index = Index(name=index_name)
        return index.put_mapping(body=mapping, using=self.elasticsearch)
    
    
    @_elastic_connection
    def add_ann_vector_mapping(
            self, 
            index_name: str,
            field: str,
            schema: Optional[dict] = None, 
            dims: int = 1024
    ) -> dict:
        vector_mapping = {
            "properties": {
                field: {
                    "type": "dense_vector", 
                    "dims": dims, 
                    "similarity": "cosine", 
                    "index": True
                }
            }
        }
        mapping = schema or vector_mapping
        index = Index(name=index_name)
        return index.put_mapping(body=mapping, using=self.elasticsearch)

    @_elastic_connection
    def add_vector(
            self, 
            index_name: str,
            document_id: str, 
            vector: List[float], 
            field: str
    ) -> dict:
        schema = {"doc": {field: vector}}
        return self.elasticsearch.update(
            index=index_name, id=document_id, body=schema, refresh="wait_for"
        )

    @_elastic_connection
    def create_index(
            self,
            index: str,
            shards: int = 3,
            replicas: int = 1,
            settings: Optional[dict] = None,
    ) -> Dict | None:
        """Creates empty index.
        :param: index str: Name of the index to create.
        :param: shards int: Number of shards for the index.
        :param: replicas int: Number of replicas of the index.
        :param: settings dict: Overwrite settings for the index.
        """

        index_exists = self.elasticsearch.indices.exists(index=index).body
        if index_exists is False:
            setting_body = settings or {
                "number_of_shards": shards,
                "number_of_replicas": replicas,
            }
            return self.elasticsearch.indices.create(index=index, settings=setting_body)

    @_elastic_connection
    def delete_index(self, index: str, ignore: Optional[bool] = True) -> Dict:
        """Deletes index.
        :param: index str: Name of the index to be deleted.
        :param: ignore bool: Ignore errors because of closed/deleted index.
        :return: Dict of Elastic's acknowledgement of the action.
        """
        response = self.elasticsearch.indices.delete(index=index, ignore_unavailable=ignore, expand_wildcards="open")
        return response

    @_elastic_connection
    def delete_document(self, index: str, document_id: str) -> ObjectApiResponse[Any]:
        """Deletes document fom index.
        :param: document_id str: ID of the document to be deleted.
        :param: index str: Index where the document is to be found.
        :param: ignore bool: Ignore errors because of closed/deleted index.
        :return: Dict of Elastic's acknowledgement of the action.
        """
        response = self.elasticsearch.delete(id=document_id, index=index)
        return response

    @_elastic_connection
    def bulk_index(
            self,
            documents: Iterator[dict],
            index_prefix: str,
            rollover_limit: int,
            refresh="false",
            create_index: bool = True
    ) -> (int, int):
        last_index_name = self._produce_rollover_index(index_prefix, rollover_limit)
        if create_index:
            response = self.create_index(index=last_index_name)
            response = self.add_mapping(index_name=last_index_name, schema=self.generate_mapping())
            pass

        actions = [{"_index": last_index_name, "_source": document} for document in documents]
        successful_count, error_count = bulk(actions=actions, client=self.elasticsearch, max_retries=3, refresh=refresh)
        return successful_count, error_count
    
    @_elastic_connection
    def bulk_index_without_rollver(
            self,
            documents: Iterator[dict],
            index: str,
            refresh="false",
    ) -> (int, int):
        actions = [{"_index": index, "_source": document} for document in documents]
        successful_count, error_count = bulk(actions=actions, client=self.elasticsearch, max_retries=3, refresh=refresh)
        return successful_count, error_count

    @_elastic_connection
    def index_document(self, index: str, body: dict, document_id: Optional[str] = None) -> Dict:
        """Indexes document.
        :param: index str: Index that document will be indexed into.
        :param: body dict: Document body.
        :param: document_id str: Optional id for the document. Is generated automatically if None.
        :return: Dict of Elastic's acknowledgement of the action.
        """
        if document_id:
            indexed = self.elasticsearch.index(index=index, id=document_id, body=body)
        else:
            indexed = self.elasticsearch.index(index=index, body=body)
        return indexed

    @_elastic_connection
    def get_documents_by_key(self, index: str, document_key: str, sort_fields=("start_page", "end_page", "sequence_nr",)):
        index = f"{index}-*"
        s = elasticsearch_dsl.Search(using=self.elasticsearch, index=index)
        s = s.query("match", parent_id=document_key).sort(*sort_fields)
        # Since scan doesn't allow for sorting, we do it manually after fetching the documents.
        documents = sorted(
            s.scan(), key=lambda doc: [getattr(doc, field) for field in sort_fields]
        )
        return documents
    
    
    @_elastic_connection
    def execute_fuzzy_search(
            self, 
            index: str,
            field: str,
            entity: str, 
            fuzziness: int = 2, 
            prefix_length: int = 1, 
            max_expansions: int = 50
    ) -> Response:
        """Executes a fuzzy search.
        :param: index str: Index to search from.
        :param: entity str: Entity to search matches for.
        :param: fuzziness int: Maximum edit distance for a match.
        :param: prefix_length int: Number of characters in the prefix that 
            should overlap with the original entity's prefix.
        :param: max_expansion int: maximum number of terms the fuzzy query 
            will match before halting the search
        :return: Dict on search results.
        """  
        query_params = {
            f"{field}.keyword": {
                "value": entity,
                "fuzziness": fuzziness,
                "max_expansions": max_expansions,
                "prefix_length": prefix_length
            }
        }
        s = elasticsearch_dsl.Search(using=self.elasticsearch, index=index)
        s = s.query("fuzzy", **query_params)
        response = s.execute()
        return response
    
    def execute_ann_vector_search(
            self,
            index: str,
            field: str,
            query_vector: List[float],
            k: int = 10,
            num_candidates: int = 100,
            n_docs: int = 10,
            elastic_ids: List[str] = []
    ) -> Response:
        """ Execute a vector search.
        NB! Works only with ANN mapping!
        
        :param: index str: Index to search from.
        :param: field str: Field containing vectorized data.
        :param: query vector List[float]: Vector to search matches for.
        :param: k int: Number of nearest neighbors to return.
        :param: num_candidates int: Number of candidates considered before selecting k results.
        :param: n_docs: int: Number of documents to return.
        :param: elastic_ids: List[str]: Elastic ID-s for restricting the search.
        """

        s = elasticsearch_dsl.Search(using=self.elasticsearch, index=index)

        # Add kNN vector search
        s = s.extra(
            knn={
                "field": field,
                "query_vector": query_vector, 
                "k": k, 
                "num_candidates": num_candidates
            }
        )

        # Add ID filtering, if elastic_ids are specified
        if elastic_ids:
            s = s.query(
                elasticsearch_dsl.Q("terms", _id=elastic_ids)
            )
            
        # Sort by score and return `n_docs` best-matching documents
        s = s.extra(size=n_docs)

        # Execute the search
        response = s.execute()
        return response
    
        
    def execute_script_score_vector_search(
            self,
            index: str,
            field: str,
            query_vector: List[float],
            n_docs: int = 10,
            elastic_ids: List[str] = []
    ) -> Response:
        """ Execute a vector search.
        NB! Requires different mapping than ANN!
        
        :param: index str: Index to search from.
        :param: field str: Field containing vectorized data.
        :param: query vector List[float]: Vector to search matches for.
        :param: n_docs: int: Number of documents to return.
        :param: elastic_ids: List[str]: Elastic ID-s for restricting the search.
        """
        s = elasticsearch_dsl.Search(using=self.elasticsearch, index=index)
        
        if elastic_ids:
            query = elasticsearch_dsl.Q("terms", _id=elastic_ids)
        else:
            query = elasticsearch_dsl.Q("match_all")
        # Apply script_score query
        s = s.query(
            "script_score",
            query=query,
            script={
                "source": f"1.0 + cosineSimilarity(params.query_vector, '{field}')",
                "params": {
                    "query_vector": query_vector
                }
            }
        )
        # Set min_score and limit number of documents
        s = s.extra(size=n_docs)

        # Execute search
        response = s.execute()
        return response


    def __str__(self) -> str:
        return self.elasticsearch_url
